#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

// Define block size
#define BLOCK_SIZE 256

// Fused CUDA kernel that performs softmax across channels followed by a sigmoid activation.
// It uses an online reduction algorithm to compute the maximum and the sum of exponentials in a single pass,
// thus reducing the number of global memory accesses and iterations over the channel dimension.

// The online algorithm works as follows:
//   Initialize: m = x[0], t = 1 (equivalent to exp(x[0]-m) = 1).
//   For each channel c from 1 to channels-1:
//     if (x[c] > m):
//         t = t * exp(m - x[c]) + 1;
//         m = x[c];
//     else:
//         t += exp(x[c] - m);
// After the loop, the softmax for a given channel is computed as:
//     softmax(x[c]) = exp(x[c]-m) / t,
// and then sigmoid is applied: sigmoid(softmax) = 1 / (1 + exp(-softmax)).

template <typename scalar_t>
__global__ void fused_softmax_sigmoid_kernel(
    const scalar_t* __restrict__ input,
    scalar_t* __restrict__ output,
    const int batch,
    const int channels,
    const int depth,
    const int height,
    const int width) {

    // Each thread processes one spatial location (pixel) over all channels
    int spatial = depth * height * width;
    int total_pixels = batch * spatial;
    int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < total_pixels) {
        // Compute batch and spatial coordinates
        int b = idx / spatial;
        int pixel_idx = idx % spatial;
        int d = pixel_idx / (height * width);
        int rem = pixel_idx % (height * width);
        int h = rem / width;
        int w = rem % width;

        // Compute base offset for this pixel across all channels
        // Data layout: [batch, channels, depth, height, width]
        int base = (b * channels * spatial) + (d * height * width + h * width + w);
        int stride = spatial; // jumping one channel

        // Use online reduction to compute softmax normalization factors in one pass
        scalar_t max_val = input[base];  // assume channels >= 1
        scalar_t sum_exp = 1.0f;  // corresponds to exp(input[base] - max_val) where max_val == input[base]
        
        for (int c = 1; c < channels; c++) {
            int pos = base + c * stride;
            scalar_t val = input[pos];
            if (val > max_val) {
                // Adjust previous sum to new max
                sum_exp = sum_exp * exp(max_val - val) + 1.0f;
                max_val = val;
            } else {
                sum_exp += exp(val - max_val);
            }
        }

        // Second pass: compute softmax and apply sigmoid activation
        for (int c = 0; c < channels; c++) {
            int pos = base + c * stride;
            scalar_t softmax_val = exp(input[pos] - max_val) / sum_exp;
            // Apply sigmoid activation to the softmax value
            output[pos] = 1.0f / (1.0f + exp(-softmax_val));
        }
    }
}


// Forward function: applies conv_transpose3d using PyTorch built-in, then the fused CUDA kernel
torch::Tensor forward(
    torch::Tensor input,
    int stride,
    int padding,
    int output_padding,
    bool bias_flag,
    torch::Tensor conv_transpose,
    torch::Tensor conv_transpose_bias) {

    // Perform the 3D transpose convolution using PyTorch's conv_transpose3d
    auto x = torch::conv_transpose3d(
        input,
        conv_transpose,
        bias_flag ? conv_transpose_bias : torch::Tensor(),
        stride,
        padding,
        output_padding
    );

    const int batch = x.size(0);
    const int channels = x.size(1);
    const int depth = x.size(2);
    const int height = x.size(3);
    const int width = x.size(4);

    auto output = torch::empty_like(x);

    int spatial = depth * height * width;
    int total_pixels = batch * spatial;
    int threads = BLOCK_SIZE;
    int blocks = (total_pixels + threads - 1) / threads;

    AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "fused_softmax_sigmoid_kernel", ([&] {
        fused_softmax_sigmoid_kernel<scalar_t><<<blocks, threads>>>(
            x.data_ptr<scalar_t>(),
            output.data_ptr<scalar_t>(),
            batch,
            channels,
            depth,
            height,
            width
        );
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused ConvTranspose3d with Softmax and Sigmoid using online reduction");
}
